/*----------------------------------------------------------------------
 *
 * PROGRAM: mqput.c
 *
 *  Author: Sid Young (sidy)
 *    Date: 2-1-2001
 *
 * Purpose: To read in a file and write it to the nominated queue.
 *  Limits: File must be less than MSG_LIMIT (default queue size per msg).
 *
 * Assumes that the MQSERVER environment variable is set to the MQ Server.
 * Default format is export MQSERVER=CHANNEL_NAME/TCP/HOST
 *
 *
 * History
 * ----------------------------------------------------------------------
 *
 * 02-01-2001  sidy  Began coding using MQ sample as base.
 * 29-10-2001        Cleaned up code
 *                   Changed PMO and OPEN options to include application ID
 *                   so we can retrieve the filename from the MQMD struct.
 * 02-02-2007        Added AES support and cmd line options
 * 06-02-2007        Changed header block to support IBM/HPUX and Linux systems
 *                   Added support for MD5 hash
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <malloc.h>

#include <cmqc.h>
#include "rijndael.h"
#include "md5.h"


MQOD  od = {MQOD_DEFAULT};    /* Object Descriptor */
MQMD  md = {MQMD_DEFAULT};    /* Message Descriptor */
MQPMO pmo= {MQPMO_DEFAULT};   /* put message options */


unsigned int dhFlag;
unsigned int aesFlag;



/*
 * header block used for AES/Rijndael encrypted messages
 *
 */
struct HDR_BLOCK
{
	unsigned char hdr_length[4];
	unsigned char hdr_major;
	unsigned char hdr_minor;
	char src_hash[34];
	unsigned char file_length[4];
	char file_name[64];
	char file_type[4];
	unsigned char hdr_flag_1;	/* flag used to hold 'E'ncrypted data block */
	unsigned char hdr_flag_2;
	unsigned char hdr_flag_3;
	unsigned char hdr_flag_4;	/* OS Flag for PUT utility */
};


struct MD5Context *md5ctx;



#define MSG_LIMIT  4*1024*1024

struct stat fileStat;
char *pSrcBuffer;	/* clear text data from file */
char *ptrDestBuffer;
char *pDestBuffer;	/* destination pointer of encryption buffer */
char *endPtr;
char *workPtr;
FILE *fPtr;
long fileSize;
char key[32];
char szBuffer[1024];
char szQueueManager[64];
char szFileName[64];
char szAESKey[64];
char szQueueName[64];
char szHash[64];
char szFileType[4];







/**
 * getHeaderLength()
 *
 *
 * Utility to return length of header
 * given pointer to header block
 *
 *
 */
long getHeaderLength( struct HDR_BLOCK *hdr )
{
static long _HdrBlockLength;

	_HdrBlockLength =hdr->hdr_length[0];
	_HdrBlockLength|=hdr->hdr_length[1]<<8;
	_HdrBlockLength|=hdr->hdr_length[2]<<16;
	_HdrBlockLength|=hdr->hdr_length[3]<<24;
	return _HdrBlockLength;
}








/**
 * getFileLength()
 *
 *
 * Helper utility to return file length stored in 
 * header block passed in message buffer.
 *
 *
 *
 *
 */
long getFileLength( struct HDR_BLOCK *hdr )
{
static long _SavedFileLength;

	_SavedFileLength= hdr->file_length[0];
	_SavedFileLength|=hdr->file_length[1]<<8;
	_SavedFileLength|=hdr->file_length[2]<<16;
	_SavedFileLength|=hdr->file_length[3]<<24;
	return _SavedFileLength;
}





/**
 *
 * AllocateAndReadFile()
 *
 * Given a file name, attempt to size it and then allocate space if it
 * is a valid file.
 *
 * @return   int 0= OK
 */
int AllocateAndReadFile(char *fileName )
{
char thisChar;

	if(!stat( fileName, &fileStat))
		fileSize = fileStat.st_size;
	else
	{
		printf("ERROR - Cannot STAT file, check spelling!\n\n");
		exit(1);
	}
	if(fileSize > MSG_LIMIT )
	{
		printf("ERROR - File must be less than 4MB\n\n");
		exit(2);
	}
	if(S_ISDIR(fileStat.st_mode))
	{
		printf("ERROR - Cannot process a directory!\n\n");
		exit(3);
	}
/*
 *  Attempt to allocate memory and open file
 */
	if((pSrcBuffer = (char*) malloc( fileSize+1)) != NULL)
		workPtr = pSrcBuffer;
	else
	{
		printf("ERROR - Cannot Allocate any memeory!\n\n");
		exit(4);
	}
	if((fPtr=fopen(fileName,"r"))!=NULL)
	{
		while(!feof(fPtr))
		{
			thisChar = fgetc(fPtr);
			if(feof(fPtr)) break;
			*workPtr++ = thisChar;
		}
		endPtr = workPtr-1;
		fclose(fPtr);
	}
	else
	{
		printf("ERROR - Cannot open file for reading!\n\n");
		free(pSrcBuffer);
		exit(5);
	}
	return 0;
}



/**
 * DumpBlock()
 *
 *
 * Dump the Header to screen if the -dh flag
 * is specified on the command line
 *
 * @pre    None
 * @post   None
 * @return void
 */
void DumpBlock(struct HDR_BLOCK *hdr)
{

	if(dhFlag==0) return;

	printf("Header Dump\n");
	printf("===========\n");
	printf("Version    [%c.%c] \n", hdr->hdr_major, hdr->hdr_minor);
	printf("Block Size [%ld]\n", getHeaderLength(hdr));
	printf("Hash       [%s]\n", hdr->src_hash);
	printf("File Name  [%s]\n", hdr->file_name);
	printf("File Length[%ld]\n", getFileLength(hdr));
	printf("FileType   [%s]\n", hdr->file_type);
	printf("Flag 1     [%c]\n", hdr->hdr_flag_1);
	printf("Flag 2     [%c]\n", hdr->hdr_flag_2);
	printf("Flag 3     [%c]\n", hdr->hdr_flag_3);
	printf("Flag 4     [%c]\n", hdr->hdr_flag_4);
}




/**
 * EncryptDataBlock()
 *
 *
 * Given a source buffer, encrypt the data to the destination buffer
 * using the "key" in the ctx structure. 
 * Encryption buffer will be large by anything up to 15 bytes more 
 * than original buffer.
 *
 *
 * NOTE: byte ordering is different on AIX/Linux
 */
long EncryptDataBlock(struct HDR_BLOCK * hdr, rijndael_ctx *ctx, char *pSource, char *pDestBuffer, long flen)
{
unsigned int blklen;    /* length of header block */
long filelen;	/* original file length */
long blkcnt;	/* number of 16 byte blocks */
long maxblks;	/* max blocks to fit all data */
long maxflen;	/* size in bytes */

	if(dhFlag==1) printf("EncryptDataBlock()\n");

	blkcnt=flen/16;     /* number of 16 byte blocks */
	maxblks=blkcnt;     /* max blocks to fit all data */
	maxflen=maxblks*16; /* size in bytes */
	if(dhFlag==1)
		printf("BLKS [%ld] LEN [%ld] MAX [%ld]\n", maxblks, flen, maxflen);
	if((blkcnt*16)!=flen)
	{
		if(dhFlag==1) printf("Adjusting blocks\n");
		maxblks++; maxflen=maxblks*16;
		if(dhFlag==1)
			printf("BLKS [%ld] LEN [%ld] MAX [%ld]\n", maxblks, flen, maxflen);
	}
/*
 *---------------------------------------------
 * Setup Header block with information about file
 * save file length in intel byte order H-L format
 *---------------------------------------------
 */
	hdr->hdr_major='1';
	hdr->hdr_minor='1';
	strcpy( hdr->file_name, szFileName);
	strcpy(hdr->file_type, szFileType);
	hdr->hdr_flag_1='E';
	hdr->hdr_flag_4='L';	/* L=Linux */
	memset(hdr->src_hash,0, sizeof(hdr->src_hash));
/* platform neutral write */
	memset(hdr->hdr_length,0,4);

	hdr->hdr_length[0] = sizeof(struct HDR_BLOCK) &0xff;
	hdr->hdr_length[1] =(sizeof(struct HDR_BLOCK)>>8 )&0xff;
	hdr->hdr_length[2] =(sizeof(struct HDR_BLOCK)>>16)&0xff;
	hdr->hdr_length[3] =(sizeof(struct HDR_BLOCK)>>24)&0xff;

	hdr->file_length[0] =flen&0xff;
	hdr->file_length[1] =(flen>>8)&0xff;
	hdr->file_length[2] =(flen>>16)&0xff;
	hdr->file_length[3] =(flen>>24)&0xff;

#ifdef _AIX_SOURCE 
	hdr->hdr_flag_4='I';	/* I=IBM */
#endif
#ifdef _HPUX_SOURCE
	hdr->hdr_flag_4='H';	/* H=HPUX */
#endif
/*
 *---------------------------------------------
 * Setup context and then encrypt blocks
 *---------------------------------------------
 */
	rijndael_set_key(ctx, key, 128, 0);
	while(maxblks--)
	{
		if(dhFlag==1) printf(".");
		rijndael_encrypt(ctx, pSource, pDestBuffer);
		pSource+=16; pDestBuffer+=16;
	}
	return maxflen;
}





/**
 * CalculateMD5()
 *
 *
 * Calculate the MD5 hash from the supplied buffer
 *
 *
 */
char * CalculateMD5(struct MD5Context *ctx, char *pSrc, long fileSize)
{
unsigned char szTmpBuffer[128];
static char szFinalBuffer[128];
unsigned char tmpSig[4];
int i;

	printf("DUMP\n");

	for(i=0;i<fileSize;i++)
		printf("%02X ", *(pSrc+i));



	memset(szFinalBuffer,0,sizeof(szFinalBuffer));
	memset(szTmpBuffer,0,sizeof(szTmpBuffer));
	MD5Init( ctx );
	MD5Update( ctx,  pSrc, fileSize);
	MD5Final( szTmpBuffer, ctx );	/* binary hash value */
	for(i=0;i<16;i++)
	{
		sprintf(tmpSig,"%02X", szTmpBuffer[i]);	/* convert to ASCII */
		strcat(szFinalBuffer, tmpSig);
	}
	return szFinalBuffer;
}







/*=================================================================
 *
 *
 *
 *
 *                      M A I N
 *
 *
 *
 *
 *
 */
int main(int argc, char **argv)
{
FILE *fp;
MQHCONN  hConn;
MQHOBJ   hObject;
MQLONG   _OpenOptions;
MQLONG   _CloseOptions;
MQLONG   _OpenCode;			/* MQOPEN completion code */
MQLONG   _CC;				/* completion code*/
MQLONG   _RC;				/* reason code */
MQLONG   _ConnRC;			/* reason code for MQCONN */
MQLONG   messlen;			/* message length */
char     buffer[100];		/* message buffer */
int i;
long   _destbuffsize;

struct HDR_BLOCK *hdr;
rijndael_ctx *ctx;


	printf("\nMQPUT v2.0 (aes) - Place a file into a Message Queue\n\n");
	printf("Build Date: %s  Time: %s\n", __DATE__, __TIME__);
	printf("Author: Sid Young\n");
	if (argc < 3)
	{
		printf("Required parameter missing:\n\n");
		printf("\t%s <queue_name> <file_to_send>\n\n\n", argv[0]);
		exit(99);
	}
/*
 *---------------------------------------------
 *       Process command line args
 *---------------------------------------------
 */
	szQueueManager[0] = 0;		/* just use default queue manager*/
	strcpy(szFileType, "???");	/* default */
	if(argc>1)
	{
		for(i=0;i<argc;++i)
		{
			if(!strcmp(argv[i],"-?"))
			{
				printf("HELP: mqput <queue_name> <file> [-key <key>] [-type <type>] [-dh]\n");
				printf(" <queue_name> - Name of Queue to put to\n");
				printf(" <file> - Name of file to send\n");
				printf(" -key  <key> - 16 byte hex string for cipher\n");
				printf(" -type <type> - file type string (3 bytes)\n");
				printf(" -dh   - dump header if encrypted\n");
				exit(0);
			}
			if(!strcmp(argv[i],"-dh")) { dhFlag=1; }
			if(!strcmp(argv[i],"-key")) {aesFlag=1;strcpy(szAESKey,argv[i+1]);}
			if(!strcmp(argv[i],"-int"))
			{
				printf("Sizeof int = %d\n", sizeof(unsigned int));
				printf("Sizeof long= %ld\n", sizeof(unsigned long));
			}
			if(!strcmp(argv[i],"-type"))
			{
				if(strlen(argv[i+1])>3)
					strncpy( szFileType, argv[i+1], 3);
				else
					strcpy( szFileType, argv[i+1]);
				szFileType[3]=0;
			}
		}
		if(dhFlag==1)
		{
			printf("Command line options:\n");
			for(i=0;i<argc;++i)
 				printf("Arg [%d] Value [%s]\n",i, argv[i]); 
		}
	}
	strcpy(od.ObjectName, argv[1]);
	strcpy(szQueueName,argv[1]);
	strcpy(szFileName, argv[2]);
/*
 *---------------------------------------------
 * Allocate a header to hold details about message
 *---------------------------------------------
 */
 	if(aesFlag==1)
	{
		printf("AES Encryption required\n");
  		if((hdr=(struct HDR_BLOCK *)malloc(sizeof(struct HDR_BLOCK)))==NULL)
		{
			printf("ERROR - Cannot allocate HDR block\n");
			exit(-1);
		}
		memset(hdr,0,sizeof(struct HDR_BLOCK));
/*
 *---------------------------------------------
 * Allocate a context buffer for crypto call.
 *---------------------------------------------
 */
  		if((ctx=(rijndael_ctx *)malloc(sizeof(rijndael_ctx)))==NULL)
		{
			printf("ERROR - Cannot allocate CTX block\n");
			exit(-1);
		}
/*
 *---------------------------------------------
 * make sure key is correct length 128 bits minimum
 *---------------------------------------------
 */
		while(strlen(szAESKey)<32)
			strcat(szAESKey, "0");
		memcpy(key,szAESKey,32);
		printf("AES Key [%s]\n",key);
	}
/*
 *---------------------------------------------
 *  Connect to queue manager
 *---------------------------------------------
 */
	szQueueManager[0] = 0;
	printf("Attempt connect -> ");
	MQCONN(szQueueManager, &hConn, &_CC, &_ConnRC); 
/*
 *---------------------------------------------
 * report reason and stop if it failed
 *---------------------------------------------
 */
	if(_CC == MQCC_FAILED)
	{
		printf("ERROR - MQCONN ended with reason code [%ld]\n", _ConnRC);
		exit( 98 );
	}
	printf("Connected to Queue Manager\n");
/*
 *---------------------------------------------
 *  Use parameter as the name of the target queue
 *---------------------------------------------
 */

   strncpy(od.ObjectName, argv[1], (size_t)MQ_Q_NAME_LENGTH);
   printf("Target queue: [%s]\n", od.ObjectName);
/*
 *---------------------------------------------
 * Open the target message queue for output,
 * but not if MQM stopping.
 *---------------------------------------------
 */
	_OpenOptions = MQOO_OUTPUT + MQOO_FAIL_IF_QUIESCING +MQOO_SET_ALL_CONTEXT;
	printf("Open queue....\n");
	MQOPEN(hConn, &od, _OpenOptions, &hObject, &_OpenCode, &_RC);
/*
 *---------------------------------------------
 * report reason, if any; stop if failed
 *---------------------------------------------
 */
	if(_RC != MQRC_NONE)
		printf("ERROR - MQOPEN ended with reason code [%ld]\n", _RC);

	if(_OpenCode == MQCC_FAILED)
		printf("ERROR - Unable to open queue for output\n");

/*
 *---------------------------------------------
 *  Read from the file and put the message buffer to the message queue
 *  Loop until null line or end of file, or there is a failure
 *---------------------------------------------
 */
	if(_RC == MQRC_NONE)
		AllocateAndReadFile( argv[2] );
/*
 *---------------------------------------------
 * Put the memory buffer and free our allocated memory
 *---------------------------------------------
 */
	_CC = _OpenCode;
	memcpy(md.Format, MQFMT_STRING, (size_t)MQ_FORMAT_LENGTH);
	pmo.Options |= MQPMO_NEW_MSG_ID;
	pmo.Options |= MQPMO_SET_IDENTITY_CONTEXT;
	memset(md.ApplIdentityData,0,sizeof(md.ApplIdentityData));
	strcpy(md.ApplIdentityData,argv[2]);
/*
 *---------------------------------------------
 * Allocate block for MD5 hash routine
 *---------------------------------------------
 */
	md5ctx = (struct MD5Context *) malloc(sizeof(struct MD5Context));
	memset(md5ctx,0,sizeof(struct MD5Context));
	printf("HASH [%s]\n",CalculateMD5( md5ctx, pSrcBuffer, fileSize));


/*
 *---------------------------------------------
 * Allocate block for encrypted data and header
 * then encrypt file into destination block.
 * Otherwise:
 * PUT buffer to queue for length of file.
 *---------------------------------------------
 */
	if(aesFlag==1)
	{
		_destbuffsize = fileSize+32+sizeof(struct HDR_BLOCK);
		if((pDestBuffer=(char*) malloc(_destbuffsize)) != NULL)
		{
			ptrDestBuffer=pDestBuffer;	/* keep original pointer */

			pDestBuffer+=sizeof(struct HDR_BLOCK);	/* add in header space */
			_destbuffsize=EncryptDataBlock(hdr,ctx, pSrcBuffer, pDestBuffer, fileSize);
			DumpBlock( hdr );
			pDestBuffer=ptrDestBuffer;	/* reset pointer */
			memcpy(ptrDestBuffer, hdr, sizeof(struct HDR_BLOCK));
			_destbuffsize+= sizeof(struct HDR_BLOCK);
			MQPUT(hConn,hObject,&md,&pmo,_destbuffsize,pDestBuffer,&_CC,&_RC);
			free(pDestBuffer);
			pDestBuffer=NULL;
			ptrDestBuffer=NULL;
		}
		else
		{
			printf("ERROR - Cannot allocate memory for encryption\n");
		}
	}
	else /* put clear text file */
	{
		MQPUT(hConn,hObject,&md,&pmo,fileSize,pSrcBuffer,&_CC,&_RC);
	}
	free(pSrcBuffer);
	pSrcBuffer=NULL;
	if(_RC != MQRC_NONE)
		printf("ERROR - MQPUT ended with reason code [%ld]\n", _RC);
/*
 *---------------------------------------------
 * Close the target queue (if it was opened)
 *---------------------------------------------
 */
	if(_OpenCode != MQCC_FAILED)
	{
		_CloseOptions = 0;
		MQCLOSE(hConn, &hObject, _CloseOptions, &_CC, &_RC);
		if(_RC != MQRC_NONE)
		printf("MQCLOSE ended with reason code [%ld]\n", _RC);
	}
/*
 *---------------------------------------------
 * Disconnect from MQM if connected
 *---------------------------------------------
 */
	if(_ConnRC != MQRC_ALREADY_CONNECTED)
	{
		MQDISC(&hConn, &_CC, &_RC);
		if(_RC != MQRC_NONE)
   		printf("ERROR - MQDISC ended with RC [%ld]\n", _RC);
	}
	printf("File Transfer Complete.\n\n");
	return(0);
}
/*
 * end of file
 */
